var sprites = [];

class Sprite {
    constructor(imgSrc, x, y) {
        this.img = new Image();
        this.img.src = imgSrc;
        this.x = x;
        this.y = y;
        this.distance = 0;
        this.angle = 0;
        this.visible = true;
        this.solid = true; // allow for ghosts maybe
    }
    getDims() {
        return { width: this.img.width, height: this.img.height };
    }
}
class Sprite3D extends Sprite{
    constructor(imgSrc, x, y, frameWidth, rot, frames) {
        super(imgSrc, x, y);
        this.rotation = rot;
        this.frame = -1;
        this.frameWidth = frameWidth;
        this.canvas = document.createElement('canvas');
        this.ctx = null;
        this.frameCount = frames;
    }
    getDims() {
        return { width: this.frameWidth, height: this.img.height };
    }
    drawFrame(newFrame) {
        if (!this.img.height) { return; } // image not loaded yet
        let width = this.frameWidth, height = this.img.height;
        if (!this.ctx) { // first use of method
            this.canvas.width = width;
            this.canvas.height = height;
            this.ctx = this.canvas.getContext("2d");
        }
        this.frame = newFrame;
        this.ctx.clearRect(0, 0, width, height);
        let imgX = this.frame * width;
        this.ctx.drawImage(this.img, imgX, 0, width, height, 0, 0, width, height);
    }
    viewedFrom(x, y) {
        let dx = x - this.x; // offset from sprite
        let dy = y - this.y;
        let angle = Math.atan2(dy, dx); // angle from sprite to player
        angle = only360(this.rotation - angle);
        let ai = getAngleIndex(getNearestAngle(angle / Math.PI));
        if (this.frame !== ai) {
            this.drawFrame(ai); // redraw to image buffer
        }
    }
}
class Anim {
    constructor(imgSrc, frames, frameWidth, height, repeat, rate) {
        this.loaded = false;
        this.img = new Image();
        let self = this;
        //inside img.onload this == img but self == Anim this
        this.img.onload = function () { self.loaded = true; };
        this.img.src = imgSrc;
        this.frames = frames;
        this.frameWidth = frameWidth;
        this.height = height;
        this.frame = -1;
        this.default = 0;
        this.repeat = repeat;
        this.frameRate = 1000 / rate;
        this.lastFrame = Date.now();
        this.action = false;
    }
    activate(yesNo){ this.action = yesNo;}
    reset() { this.frame = this.default; }
    getDims() {
        return { width: this.frameWidth, height: this.height };
    }
    animate(ctx, x, y) {
        let now = Date.now();
        if (this.action) {
            if ((now - this.lastFrame) >= this.frameRate) {
                this.frame++;
                this.lastFrame = now;
                if (this.frame === this.frames) {
                    if (this.repeat) {
                        this.frame = 1;
                    } else {
                        this.frame = this.default;
                        this.action = false;
                    }
                }
            }
        } else {
            this.frame = this.default;
        }
        let clipx = this.frame * this.frameWidth;
        ctx.drawImage(this.img, clipx, 0, this.frameWidth, this.height, x, y, this.frameWidth, this.height);
    }
}
class AnimAudio extends Anim {
    constructor(imgSrc, frames, frameWidth, height, repeat, rate, audSrc){
        super(imgSrc, frames, frameWidth, height, repeat, rate);
        this.sound = new Audio();
        this.sound.src = audSrc;
        this.sound.onended = function(){
            this.currentTime = 0;
        };
    }
    activate(yesNo) {
        super.activate(yesNo); //this.action = yesNo; just as valid
        if(yesNo){this.sound.play();}
    }
}
class Actor {
    constructor(x, y) {
        this.x = x;
        this.y = y;
        this.anims = [];
        this.current = -1;
        this.height = 0;
        this.width = 0;
        this.visible = true;
        this.solid = true;
    }
    resetCurrent(newc) {
        this.current = newc;
        this.anims[this.current].reset();
        let dim = this.anims[this.current].getDims();
        this.width = dim.width;
        this.height = dim.height;
    }
    activate() {
        if (this.anims[this.current] instanceof AnimAudio) {
            if (this.anims[this.current].sound.currentTime) { return; }
        }
        this.anims[this.current].activate(true);
    }
}
class NPC extends Actor{
    constructor(x, y, rot, prog) {
        super(x,y);
        this.turn = 0;
        this.rotation = rot;
        this.speed = 0;
        this.mapRadius = 3;
        this.angle = 0;
        this.distance = 0;
        this.visible = true;
        this.canvas = document.createElement('canvas');
        this.ctx = null;
        this.manageActions = prog;
        this.state = NPC.states.idle;
        this.playerView = {};
        this.blocked = false;
        this.active = false;
        this.addAnims(); //
    }
    addAnims() {
        let a = new Anim("images/guard0x5x42.png", 5, 42, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard1x5x44.png", 5, 44, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard2x5x46.png", 5, 46, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard3x5x42.png", 5, 42, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard4x5x42.png", 5, 42, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard5x5x46.png", 5, 46, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard6x5x46.png", 5, 46, 64, true, 8);
        this.anims.push(a);
        a = new Anim("images/guard7x5x46.png", 5, 46, 64, true, 8);
        this.anims.push(a);
        a = new AnimAudio("images/guard8x8x33.png", 8, 33, 64, true, 8, "audio/guardgun.wav");
        this.anims.push(a);
        a = new Anim("images/guard9x8x62.png", 8, 62, 64, false, 8);
        this.anims.push(a);
    }
    /* jshint ignore:start */
    static states = {"idle": 1, "seen":2, "active": 4,"stand" : 8, "fight": 16, "dead": 32};
    /* jshint ignore:end */
    getDims() {
        return { width: this.width, height: this.height };
    }
    animate() {
        if (this.turn) {
            this.rotation = only360(this.rotation + this.turn);
            this.turn = 0;
        }
        if (this.speed) {
            let newX = this.x + Math.cos(this.rotation) * this.speed;
            let newY = this.y + Math.sin(this.rotation) * this.speed;
            let pos = detectCollision3(this.x, this.y, newX, newY, this.width / 2, this.rotation, 1);
            //let pos = detectCollision2(this.x, this.y, newX, newY, this.width / 2);
            if (pos.x === this.x && pos.y === this.y) {
                this.blocked = true; // if blocked stop activity
                if (this.current > -1) {
                    this.anims[this.current].activate(false);
                }
                this.active = false;
            } else {
                this.x = pos.x;
                this.y = pos.y;
            }
            this.speed = 0;
        }
        if (this.current > -1) {
            this.ctx.clearRect(0, 0, this.width, this.height);
            this.anims[this.current].animate(this.ctx, 0, 0);
        }
        if (this.manageActions) { this.manageActions(this); }
    }
    viewedFrom(x, y) {
        let dx = x - this.x;
        let dy = y - this.y;
        this.playerView.angle = only360(Math.atan2(dy, dx));
        this.playerView.index = getAngleIndex4(this.playerView.angle / Math.PI);
        this.playerView.distance = Math.hypot(dx, dy);
        let angle = only360(this.rotation - this.playerView.angle);
        let ai = getAngleIndex(getNearestAngle(angle / Math.PI));
        if (ai != this.current && this.current <= 7) { // ***
            if (!this.anims[ai].loaded) { return; }
            this.resetCurrent(ai);
            if (this.active) {
                this.anims[this.current].activate(true);
            }
        }
    }
    resetCurrent(newc){
        super.resetCurrent(newc);
        this.setCanvas();
    }
    setCanvas() {
        this.canvas.width = this.width;
        this.canvas.height = this.height;
        if (!this.ctx) { this.ctx = this.canvas.getContext("2d"); }
    }
    activate(yesNo) {
        this.active = yesNo;
        if (this.current > -1) {
            this.anims[this.current].activate(yesNo);
        }
    }
    howFar() {
        // gets distances to wall in absolute directions (0 = right on map)
        let res = {};
        for (let i = 0; i < 4; i++) {
            res['d' + i] = camera.trackRay(i * Math.PI / 2, -1, this.x, this.y);
        }
        return res;
    }
    static getBest(ds, idx){
        let best = 0, newDir = 0;
        for(let i = 0; i < 4; i++){
            if(ds['d' + i] > best && i !== idx){
                best = ds['d' + i];
                newDir = i;
            }
        }
        return newDir;
    }
    static readyToWalk(self){
        self.manageActions = NPC.doWalk;
        self.state |= NPC.states.active;
    }
    static doWalkOrig(self){
        if(!(self.state & NPC.states.seen) && !(self.state & NPC.states.active)){return;}
        if(self.rotation === 0){
            self.turn = Math.PI / 2;
            self.activate(true);
        }
        if(self.blocked){
            self.turn = Math.PI;
            self.blocked = false;
            self.activate(true);
        } else {self.speed = 1.5;}
    }
    static doWalk(self){
        if(!(self.state & NPC.states.seen) && !(self.state & NPC.states.active)){return;}
        if(self.state & NPC.states.idle){
            if(!(self.state & NPC.states.seen)){self.playerView.index = 99;} //best way
            let newDir = NPC.getBest(self.howFar(), self.playerView.index);
            self.state ^= NPC.states.idle;
            self.state ^= NPC.states.seen;
            self.state |= NPC.states.active;
            self.rotation = newDir * Math.PI / 2;
            self.activate(true);
        }
        if((self.state & NPC.states.seen) && self.playerView.distance <= 3 * Game.gameScale){
            self.manageActions = NPC.standReady;
        }
        if(self.blocked){
            self.turn = Math.PI;
            self.blocked = false;
            self.activate(true);
        } else {self.speed = 1.5;}
    }
    static standReady(self){
        if(!(self.state & NPC.states.seen)){return;}
        self.blocked = false;
        self.rotation = self.playerView.angle; // ?????????
        if(self.playerView.distance < Game.gameScale * 3){
            self.manageActions = NPC.attack;
        }
    }
    static passOn(self){
        // sadly we have a death scene
        if(!(self.state & NPC.states.dead)){
            self.resetCurrent(9);
            self.activate(true);
            self.solid = false;
            self.state |= NPC.states.dead;
            self.manageActions = null;
        }
        }
        static runAway(self) {
        if(!(self.state & NPC.states.seen) && !(self.state & NPC.states.active)){return;}
        if(self.state & NPC.states.idle){ 
            let newDir = NPC.getBest(self.howFar(), self.playerView.index);
            self.state ^= NPC.states.idle;
            self.state |= NPC.states.active;
            self.rotation = newDir * Math.PI / 2;
            self.activate(true);
        }
        //alternative is to check for new direction every grid square (avoids cul de sac)
        if(self.blocked){
            let dir = getAngleIndex4(self.rotation/Math.PI);
            let halfRad = self.width/4; // to back off a bit
            switch(dir) {
                case 0:
                case 2:
                    self.x += (dir === 0) ? -halfRad : halfRad;
                    break;
                default:
                    self.y += (dir === 1) ? -halfRad : halfRad;
                    break;
            }
            dir = (dir + 2) % 4; // 0=>2,1=>3,2=>0,3=>1
            let ds = self.howFar();
            let newDir = NPC.getBest(ds, dir); // but is best good enough?
            if(ds['d'+newDir] >= 3 * Game.gameScale){
                self.rotation = newDir * Math.PI / 2;
                self.blocked = false;
                self.activate(true);
            } else {
                self.activate(false);
                self.state ^= NPC.states.seen; 
                self.rotation = dir * Math.PI / 2;       
                self.manageActions = NPC.standReady;
            }
        } else {
            self.speed = 1.5;
        }
    }
    static attack(self) {
        if(!(self.state & NPC.states.fight)){
            self.resetCurrent(8);
            self.activate(true);
            self.state |= NPC.states.fight;
        }
        self.rotation = self.playerView.angle;
        if(self.playerView.distance > Game.gameScale * 3){
            self.activate(false);
            self.state ^= NPC.states.fight;
            self.manageActions = NPC.standReady;
        } else if(self.anims[8] instanceof AnimAudio){
            if(self.anims[8].sound.currentTime === 0){
                self.anims[8].sound.play();
                self.anims[8].action = true;
            }
        }
    }
}

// function chasePlayer(){
//     //wait to be seen and then move towards player
//     // when in range attack
// }